import numpy as np

from math import log

import warnings
warnings.filterwarnings("ignore")

# 
# By using this file, you are agreeing to this product's EULA
#
# This product can be obtained in https://github.com/jespb/Python-M6GP
#
# Copyright ©2023-2024 J. E. Batista
#

class Node:
	branches = None
	value = None


	def __init__(self):
		pass


	def create(self, rng, operators=None, terminals=None, depth=None,full=False):
		if depth>1 and (rng.random()<0.5 or full ==True ):
			op, n_args = operators[rng.randint(0,len(operators)-1)]
			self.value = op

			self.branches = []
			for i in range(n_args):
				n = Node()
				n.create(rng, operators, terminals, depth-1)
				self.branches.append(n)
		else:
			self.value = terminals[rng.randint(0,len(terminals)-1)] # Sem literais


	def copy(self,value=None, branches=None):
		self.branches = branches
		self.value=value


	def __str__(self):
		if self.branches == None:
			return str(self.value)
		else:
			if len(self.branches) == 2:
				return "( " + str(self.branches[0]) + " " + str(self.value) + " " + str(self.branches[1]) + " )"
			else:
				return str(self.value) + " ( " + " ".join( [str(b) for b in self.branches] ) + " )"


	def getSize(self):
		'''
		Returns the total number of nodes within this Node.
		'''
		if self.branches == None:
			return 1
		else:

			return 1 + sum( [b.getSize() for b in self.branches] )


	def getDepth(self):
		'''
		Returns the depth of this Node.
		'''
		if self.branches == None:
			return 1
		else:
			return 1 + max( [b.getDepth() for b in self.branches] )


	def getRandomNode(self, rng, value=None):
		'''
		Returns a random Node within this Node.
		'''
		if value == None:
			value = rng.randint(0,self.getSize()-1)
		if value == 0:
			#print(self)
			return self

		#print(value, self)
		for i in range(len(self.branches)):
			size = self.branches[i].getSize()
			if value-1 < size:
				return self.branches[i].getRandomNode(rng, value-1)
			value -= size


	def swap(self, other):
		'''
		Swaps the content of two nodes.
		'''
		b = self.branches
		v = self.value

		self.branches = other.branches
		self.value = other.value

		other.branches = b
		other.value = v


	def clone(self):
		'''
		Returns a clone of this node.
		'''
		if self.branches == None:
			n = Node()
			n.copy(value=self.value, branches = None)
			return n
		else:
			n = Node()
			n.copy(value=self.value, branches=[b.clone() for b in self.branches])
			return n



	def calculate(self, sample):
		'''
		Returns the calculated value of a sample.
		'''
		if self.branches == None:
			try:
				return np.array( sample[self.value] )#.astype("float64")
			except:
				return np.array( [float(self.value)]*sample.shape[0] )

				
		else:
			if self.value == "+": #+
				return self.branches[0].calculate(sample) + self.branches[1].calculate(sample)

			elif self.value == "-": #-
				return self.branches[0].calculate(sample) - self.branches[1].calculate(sample)

			elif self.value == "*": #*
				return self.branches[0].calculate(sample) * self.branches[1].calculate(sample)

			elif self.value == "/": #/
				right = self.branches[1].calculate(sample)
				right = np.where(right==0, 1, right)
				return self.branches[0].calculate(sample) / right

			elif self.value == "log2": # log2(X)
				res = self.branches[0].calculate(sample)
				res = np.where(res<=0, res, np.log2(res))
				return res

			elif self.value == "max": # max( X0, X1, ... Xn)
				calc = [b.calculate(sample) for b in self.branches]
				a = []
				for i in range(len(calc[0])):
					a.append( max([calc[k][i] for k in range(len(calc))]) )
				return np.array(a)

			else:
				assert False ("[ERROR] Node.calculate(self,value) > Operator %s not implemented." %self.value)
				

	def isLeaf(self):
		'''
		Returns True if the Node had no sub-nodes.
		'''
		return self.branches == None

	def getSemantics(self,tr_x):
		'''
		Returns the semantic of a Node.
		'''		
		return self.calculate(tr_x)

	def redirect(self, other):
		'''
		Assigns the content of another Node to this Node.
		'''
		self.value = other.value
		self.branches = other.branches

	def prun(self, tr_x):
		'''
		Simplifies this Node
		'''
		semantics = self.getSemantics(tr_x)
		semantics.sort()
		if semantics[0] == semantics[-1] and len(semantics)>1:
			self.value = str(semantics[0])
			self.branches = None


		
		if self.branches!=None and len(self.branches)==2: # [+, -, *, /]
			# +
			if self.value == "+":
				# 0 + X == X
				if not self.isLeaf() and ( self.branches[0].isLeaf() and self.branches[0].value == "0.0" ):
					self.redirect(self.branches[1])

				# X + 0 == X
				if not self.isLeaf() and ( self.branches[1].isLeaf() and self.branches[1].value == "0.0" ):
					self.redirect(self.branches[0])

				# X + X == 2 * X
				if not self.isLeaf() and ( str(self.branches[1]) == str(self.branches[0]) ):
					self.value = "*"
					n = Node()
					n.copy(value = "2.0")
					self.branches[0].redirect( n )

			# - 
			if self.value == "-":
				# X - 0 == X
				if not self.isLeaf() and ( self.branches[1].isLeaf() and self.branches[1].value == "0.0" ):
					self.redirect(self.branches[0])

				# X - X == 0
				if not self.isLeaf() and ( str(self.branches[1]) == str(self.branches[0]) ):
					n = Node()
					n.copy(value = "0.0")
					self.redirect( n )

			# * 
			if self.value == "*":
				# X * 0 == 0,  0 * X == 0
				if not self.isLeaf() and ( (self.branches[0].isLeaf() and self.branches[0].value=="0.0") or (self.branches[1].isLeaf() and self.branches[1].value=="0.0") ):
					n = Node()
					n.copy(value = "0.0")
					self.redirect( n )

				# 1 * X == X
				if not self.isLeaf() and ( self.branches[0].isLeaf() and self.branches[0].value == "1.0" ):
					self.redirect(self.branches[1])

				# X * 1 == X
				if not self.isLeaf() and ( self.branches[1].isLeaf() and self.branches[1].value == "1.0" ):
					self.redirect(self.branches[0])

			# //
			if self.value == "/":
				# X // 0 == 1
				if not self.isLeaf() and ( self.branches[1].isLeaf() and self.branches[1].value=="0.0" ):
					n = Node()
					n.copy(value = "1.0")
					self.redirect( n )

				# X // 1 == X
				if not self.isLeaf() and ( self.branches[1].isLeaf() and self.branches[1].value=="1.0" ):
					self.redirect(self.branches[0])

				# X // X == 1
				if not self.isLeaf() and ( str(self.branches[1]) == str(self.branches[0]) ):
					n = Node()
					n.copy(value = "1.0")
					self.redirect( n )



		if self.branches != None:
			for branch in self.branches:
				branch.prun(tr_x)


from .Individual import Individual
from .Node import Node

# 
# By using this file, you are agreeing to this product's EULA
#
# This product can be obtained in https://github.com/jespb/Python-M6GP
#
# Copyright ©2023-2024 J. E. Batista
#


def tournament(rng, population,n):
	'''
	Selects "n" Individuals from the population and return a 
	single Individual.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	candidates = [rng.randint(0,len(population)-1) for i in range(n)]
	return population[min(candidates)]

def doubleTournament(rng, population, n):
	candidate = [ tournament(rng, population, n) for i in range(n)]
	selected = candidate[0]
	for i in range(1,n):
		if candidate[i].crowdingDistance > selected.crowdingDistance:
			selected= candidate[i]
	return selected


def splitIntoFrontiers(population):
	frontiers = []
	population = population[:]

	while len(population) > 0:
		pf = getParetoFrontier(population)
		for ind in pf:
			population.pop( population.index(ind) )
		frontiers.append(pf)

	return frontiers


def getParetoFrontier(population):
	#print("  >>> Calculating Pareto Frontier")
	frontier = []
	for ind in population:

		isDominated = False
		i = 0
		while i < len(population) and not isDominated:
			isDominated =  population[i] > ind
			i = i+1
		if not isDominated:
			frontier.append(ind)
	return frontier


def getOffspring(rng, population, tournament_size):
	'''
	Genetic Operator: Selects a genetic operator and returns a list with the 
	offspring Individuals. The crossover GOs return two Individuals and the
	mutation GO returns one individual. Individuals over the LIMIT_DEPTH are 
	then excluded, making it possible for this method to return an empty list.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	#print("  >>> Breeding")

	isCross = rng.random()<0.5
	desc = None

	availableXO = [0,1]
	availableMT = [0,1,2] 

	if isCross:
		whichXO = availableXO[ rng.randint(0,len(availableXO)-1 ) ]
		if whichXO == 0:
			desc = STXO(rng, population, tournament_size)
		elif whichXO == 1:
			desc = M5XO(rng, population, tournament_size)
	else:
		whichMut = availableMT[ rng.randint(0,len(availableMT)-1 ) ]
		if whichMut == 0:
			desc = STMUT(rng, population, tournament_size)
		elif whichMut == 1:
			desc = M5ADD(rng, population, tournament_size)
		elif whichMut == 2:
			desc = M5REM(rng, population, tournament_size)
	return desc


def discardIndividualWithInvalidSize(population, min_dim, max_dim, depth_limit):
	ret = []
	for ind in population:
		dimensions = ind.getNumberOfDimensions()
		depth = ind.getDepth() if dimensions > 0 else -1
		if depth <= depth_limit and dimensions >= min_dim and dimensions <= max_dim:
			ret.append(ind)
	return ret


def STXO(rng, population, tournament_size):
	'''
	Randomly selects one node from each of two individuals; swaps the node and
	sub-nodes; and returns the two new Individuals as the offspring.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	ind1 = doubleTournament(rng, population, tournament_size)
	ind2 = doubleTournament(rng, population, tournament_size)

	d1 = ind1.getDimensions()
	d2 = ind2.getDimensions()

	r1 = rng.randint(0,len(d1)-1)
	r2 = rng.randint(0,len(d2)-1)

	n1 = d1[r1].getRandomNode(rng)
	n2 = d2[r2].getRandomNode(rng)

	n1.swap(n2)

	ret = []
	for d in [d1,d2]:
		i = Individual(ind1.operators, ind1.terminals, ind1.max_depth, ind1.model_name, ind1.fitnesses)
		i.copy(d)
		ret.append(i)
	return ret

def M5XO(rng, population, tournament_size):
	'''
	Randomly selects one dimension from each of two individuals; swaps the 
	dimensions; and returns the two new Individuals as the offspring.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	ind1 = doubleTournament(rng, population, tournament_size)
	ind2 = doubleTournament(rng, population, tournament_size)

	d1 = ind1.getDimensions()
	d2 = ind2.getDimensions()

	r1 = rng.randint(0,len(d1)-1)
	r2 = rng.randint(0,len(d2)-1)

	d1.append(d2[r2])
	d2.append(d1[r1])
	d1.pop(r1)
	d2.pop(r2)

	ret = []
	for d in [d1,d2]:
		i = Individual(ind1.operators, ind1.terminals, ind1.max_depth, ind1.model_name, ind1.fitnesses)
		i.copy(d)
		ret.append(i)
	return ret

def STMUT(rng, population, tournament_size):
	'''
	Randomly selects one node from a single individual; swaps the node with a 
	new, node generated using Grow; and returns the new Individual as the offspring.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	ind1 = doubleTournament(rng, population, tournament_size)
	d1 = ind1.getDimensions()
	r1 = rng.randint(0,len(d1)-1)
	n1 = d1[r1].getRandomNode(rng)
	n = Node()
	n.create(rng, ind1.operators, ind1.terminals, ind1.max_depth)
	n1.swap(n)


	ret = []
	i = Individual(ind1.operators, ind1.terminals, ind1.max_depth, ind1.model_name, ind1.fitnesses)
	i.copy(d1)
	ret.append(i)
	return ret

def M5ADD(rng, population, tournament_size):
	'''
	Randomly generates a new node using Grow; this node is added to the list of
	dimensions; the new Individual is returned as the offspring.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	ind1 = doubleTournament(rng, population, tournament_size)
	ret = []

	d1 = ind1.getDimensions()
	n = Node()
	n.create(rng, ind1.operators, ind1.terminals, ind1.max_depth)
	d1.append(n)

	i = Individual(ind1.operators, ind1.terminals, ind1.max_depth, ind1.model_name, ind1.fitnesses)
	i.copy(d1)
	ret.append(i)

	return ret

def M5REM(rng, population, tournament_size):
	'''
	Randomly selects one dimensions from a single individual; that dimensions is
	removed; the new Individual is returned as the offspring.

	Parameters:
	population (list): A list of Individuals, sorted from best to worse.
	'''
	ind1 = doubleTournament(rng, population, tournament_size)
	ret = []

	d1 = ind1.getDimensions()
	r1 = rng.randint(0,len(d1)-1)
	d1.pop(r1)
		
	i = Individual(ind1.operators, ind1.terminals, ind1.max_depth, ind1.model_name, ind1.fitnesses)
	i.copy(d1)
	ret.append(i)
	
	return ret



import numpy as np

# 
# By using this file, you are agreeing to this product's EULA
#
# This product can be obtained in https://github.com/jespb/Python-M6GP
#
# Copyright ©2023-2024 J. E. Batista
#

#
#   This classifier receives panda DataFrames and converts it to standard types
# to avoid "Infinite" values in the matrices.
#

def getInverseCovarianceMatrix(cluster):
	'''
	Returns the inverse covariance matrix, obtained from a cluster
	'''
	ret = []
	for i in range(len(cluster[0])):
		ret.append([0]*len(cluster[0]))
	for d1 in range(len(cluster[0])):
		for d2 in range(len(cluster[0])):
			for x in range(len(cluster)):
				ret[d1][d2] += cluster[x][d1]*cluster[x][d2]/len(cluster)
	return inverseMatrix(ret)
	
def mahalanobisDistance(v1,v2,invCovarianceMatrix):
	'''
	Returns the mahalanobis distance between two points
	'''
	if invCovarianceMatrix is None:
		return euclideanDistance(v1,v2)

	x = np.array(v1)
	y = np.array(v2)

	sub = np.subtract(x,y)
	mult = np.matmul(sub,invCovarianceMatrix)
	if (len(invCovarianceMatrix)==1):
		mult = [mult]
	mult2 = np.matmul(mult,sub.transpose())

	if(mult2 < 0):
		#print("mult2 < 0")
		return euclideanDistance(v1,v2)

	return mult2**0.5

def euclideanDistance(v1,v2):
	return sum([(v1[i]-v2[i])**2 for i in range(len(v1))])**0.5


def inverseMatrix(m):
	'''
	Returns the inverse of the matrix m, if possible,
	otherwise returns the diagonal matrix
	'''
	try:
		return np.linalg.inv(np.array(m))
	except:
		return None



class MahalanobisDistanceClassifier:

	invCovarianceMatrix = None
	classCentroids = None
	classes = None

	def __init__(self):
		pass

	def fit(self,X,Y):
		'''
		Calculates the class clusters in the output space.
		'''
		if self.invCovarianceMatrix != None:
			return

		X = [ list(sample) for sample in X.iloc ]
		Y = list(Y)

		self.classes = list(set(Y))
		clusters = []
		for i in self.classes:
			clusters.append([])

		for sample_index in range(len(X)):
			index = self.classes.index(Y[sample_index])
			coor = X[sample_index]
			clusters[index].append(coor)

		self.invCovarianceMatrix = []
		for cluster in clusters:
			m = getInverseCovarianceMatrix(cluster)
			self.invCovarianceMatrix.append(m)

		self.classCentroids = []
		for cluster in clusters:
			self.classCentroids.append([0 for i in range(len(cluster[0]))])
			for sample in cluster:
				self.classCentroids[-1] = [self.classCentroids[-1][i] + sample[i]/len(cluster) for i in range(len(self.classCentroids[-1]))]



	def predict(self, X):		
		predictions = []
		X = [ list(sample) for sample in X.iloc ]

		for sample in X:
			pick_d = mahalanobisDistance(sample, self.classCentroids[0],self.invCovarianceMatrix[0])
			pick = self.classes[0]
		
			for i in range(len(self.classes)):
				d = mahalanobisDistance(sample, self.classCentroids[i],self.invCovarianceMatrix[i])
				if d < pick_d:
					pick_d = d
					pick = self.classes[i]
		
			predictions.append(pick)
		return predictions


from .Individual import Individual
from .GeneticOperators import splitIntoFrontiers, getParetoFrontier, getOffspring, discardIndividualWithInvalidSize
import multiprocessing as mp
import time
import matplotlib.pyplot as plt

from random import Random

# 
# By using this file, you are agreeing to this product's EULA
#
# This product can be obtained in https://github.com/jespb/Python-M6GP
#
# Copyright ©2023-2024 J. E. Batista
#


class ClassifierNotTrainedError(Exception):
    """ You tried to use the classifier before training it. """

    def __init__(self, expression, message = ""):
        self.expression = expression
        self.message = message


class M6GP:

	## __INIT__ arguments
	operators = None
	max_initial_depth = None
	population_size = None
	threads: int = None
	random_state: int = 42
	rng = None # random number generator

	max_depth: int = None
	max_generation: int = None
	dim_min: int = None
	dim_max: int = None

	model_name = None 
	fitnesses = None

	verbose = None


	## FIT arguments
	terminals = None

	elite = None
	population = None
	currentGeneration = 0
	bestIndividual: Individual = None

	trainingAccuracyOverTime = None
	testAccuracyOverTime = None
	trainingWaFOverTime = None
	testWaFOverTime = None
	trainingKappaOverTime = None
	testKappaOverTime = None
	trainingMSEOverTime = None
	testMSEOverTime = None
	sizeOverTime = None
	dimensionsOverTime = None
	generationTimes = None
	solutionsPerGeneration = None



	def checkIfTrained(self):
		if self.population == None:
			raise ClassifierNotTrainedError("The classifier must be trained using the fit(Tr_X, Tr_Y) method before being used.")



	def __init__(self, operators=[("+",2),("-",2),("*",2),("/",2)], max_initial_depth = 6, 
		population_size = 500, max_generation = 100, max_depth = 17, tournament_size = 5,
		dim_min = 1, dim_max = 9999, threads=1, random_state = 42, verbose = True, 
		model_name="RandomForestClassifier", fitnesses=["Accuracy","Size"]):

		if sum( [0 if op in [("+",2),("-",2),("*",2),("/",2)] else 0 for op in operators ] ) > 0:
			print( "[Warning] Some of the following operators may not be supported:", operators)

		self.operators = operators

		self.max_initial_depth = max_initial_depth
		self.population_size = population_size
		self.threads = max(1, threads)
		self.random_state = random_state
		self.rng = Random(random_state)
		self.tournament_size = tournament_size

		self.max_depth = max_depth
		self.max_generation = max_generation
		self.dim_min = max(1, dim_min)
		self.dim_max = max(1, dim_max)

		self.model_name = model_name
		self.fitnesses = fitnesses

		self.verbose = verbose





	def __str__(self):
		self.checkIfTrained()
		return str(self.getBestIndividual())
		



	def getCurrentGeneration(self):
		return self.currentGeneration


	def getBestIndividual(self):
		'''
		Returns the final M6GP model.
		'''
		self.checkIfTrained()

		return self.bestIndividual

	def getAccuracyOverTime(self):
		'''
		Returns the training and test accuracy of the best model in each generation.
		'''
		self.checkIfTrained()

		return [self.trainingAccuracyOverTime, self.testAccuracyOverTime]

	def getWaFOverTime(self):
		'''
		Returns the training and test WAF of the best model in each generation.
		'''
		self.checkIfTrained()

		return [self.trainingWaFOverTime, self.testWaFOverTime]

	def getKappaOverTime(self):
		'''
		Returns the training and test kappa values of the best model in each generation.
		'''
		self.checkIfTrained()

		return [self.trainingKappaOverTime, self.testKappaOverTime]

	def getMSEOverTime(self):
		'''
		Returns the training and test mean squared error values of the best model in each generation.
		'''
		self.checkIfTrained()

		return [self.trainingMSEOverTime, self.testMSEOverTime]

	def getSizesOverTime(self):
		'''
		Returns the size and number of dimensions of the best model in each generation.
		'''
		self.checkIfTrained()

		return [self.sizeOverTime, self.dimensionsOverTime]

	def getGenerationTimes(self):
		'''
		Returns the time spent in each generation.
		'''
		self.checkIfTrained()

		return self.generationTimes










	def fit(self,Tr_x, Tr_y, Te_x = None, Te_y = None):
		if self.verbose:
			print("  > Parameters")
			print("    > Random State:       "+str(self.random_state))
			print("    > Operators:          "+str(self.operators))
			print("    > Population Size:    "+str(self.population_size))
			print("    > Max Generation:     "+str(self.max_generation))
			print("    > Max Initial Depth:  "+str(self.max_initial_depth))
			print("    > Max Depth:          "+str(self.max_depth))
			print("    > Minimum Dimensions: "+str(self.dim_min))
			print("    > Maximum Dimensions: "+str(self.dim_max))
			print("    > Wrapped Model:      "+self.model_name)
			print("    > Fitnesses:          "+" - ".join(self.fitnesses))
			print("    > Threads:            "+str(self.threads))
			print()

		self.Tr_x = Tr_x
		self.Tr_y = Tr_y
		self.Te_x = Te_x
		self.Te_y = Te_y
		self.terminals = list(Tr_x.columns)


		self.elite = []
		self.population = []

		while len(self.population) < self.population_size:
			ind = Individual(self.operators, self.terminals, self.max_depth, self.model_name, self.fitnesses)
			ind.create(self.rng, n_dims = self.dim_min)
			self.population.append(ind)

		self.bestIndividual = self.population[0]
		self.bestIndividual.fit(self.Tr_x, self.Tr_y)

		if not self.Te_x is None:
			self.trainingAccuracyOverTime = []
			self.testAccuracyOverTime = []
			self.trainingWaFOverTime = []
			self.testWaFOverTime = []
			self.trainingKappaOverTime = []
			self.testKappaOverTime = []
			self.trainingMSEOverTime = []
			self.testMSEOverTime = []
			self.sizeOverTime = []
			self.dimensionsOverTime = []
			self.generationTimes = []
			self.solutionsPerGeneration = []



		'''
		Training loop for the algorithm.
		'''
		if self.verbose:
			print("  > Running log:\n")

		while self.currentGeneration < self.max_generation:
			if not self.stoppingCriteria():
				t1 = time.time()
				self.nextGeneration()
				t2 = time.time()
				duration = t2-t1
			else:
				duration = 0
			self.currentGeneration += 1
			
			if not self.Te_x is None:
				if self.fitnesses[0] in ["Accuracy", "2FOLD", "WAF"]:
					self.trainingAccuracyOverTime.append(self.bestIndividual.getAccuracy(self.Tr_x, self.Tr_y, pred="Tr"))
					self.testAccuracyOverTime.append(self.bestIndividual.getAccuracy(self.Te_x, self.Te_y, pred="Te"))
					self.trainingWaFOverTime.append(self.bestIndividual.getWaF(self.Tr_x, self.Tr_y, pred="Tr"))
					self.testWaFOverTime.append(self.bestIndividual.getWaF(self.Te_x, self.Te_y, pred="Te"))
					self.trainingKappaOverTime.append(self.bestIndividual.getKappa(self.Tr_x, self.Tr_y, pred="Tr"))
					self.testKappaOverTime.append(self.bestIndividual.getKappa(self.Te_x, self.Te_y, pred="Te"))
					self.trainingMSEOverTime.append(0)
					self.testMSEOverTime.append(0)
				elif self.fitnesses[0] in ["MSE"]:
					self.trainingAccuracyOverTime.append(0)
					self.testAccuracyOverTime.append(0)
					self.trainingWaFOverTime.append(0)
					self.testWaFOverTime.append(0)
					self.trainingKappaOverTime.append(0)
					self.testKappaOverTime.append(0)
					self.trainingMSEOverTime.append(self.bestIndividual.getMSE(self.Tr_x, self.Tr_y, pred="Tr"))
					self.testMSEOverTime.append(self.bestIndividual.getMSE(self.Te_x, self.Te_y, pred="Te"))
				self.sizeOverTime.append(self.bestIndividual.getSize())
				self.dimensionsOverTime.append(self.bestIndividual.getNumberOfDimensions())
				self.generationTimes.append(duration)
				self.solutionsPerGeneration.append( [ind.fitness for ind in self.old_population] )


		# prun the final individual
		#self.getBestIndividual().prun(min_dim = self.dim_min, simp=True)




	def stoppingCriteria(self):
		'''
		Returns True if the stopping criteria was reached.
		'''
		genLimit = self.currentGeneration >= self.max_generation
		#perfectTraining = self.bestIndividual.getFitness() == 1
		
		return genLimit  #or perfectTraining




	def nextGeneration(self):
		'''
		Generation algorithm: the population is sorted; the best individual is pruned;
		the elite is selected; and the offspring are created.
		'''
		begin = time.time()

		# Calculates the accuracy of the population using multiprocessing

		#print("  >> Calculating fitness... ")
		if self.threads > 1:
			#print("Using multiprocessing")
			with mp.Pool(processes= self.threads) as pool:
				results = pool.map(fitIndividuals, [(ind, self.Tr_x, self.Tr_y) for ind in self.population] )
				for i in range(len(self.population)):
					self.population[i].trainingPredictions = results[i][0]
					self.population[i].fitness = results[i][1]
					self.population[i].training_X = self.Tr_x
					self.population[i].training_Y = self.Tr_y
					self.population[i].model = results[i][2]
		else:
			#print("Using a single thread")
			[ ind.fit(self.Tr_x, self.Tr_y) for ind in self.population]
			[ ind.getFitnesses() for ind in self.population ]

		#print("  >> Adding elite to population...")
		self.population.extend(self.elite)

		#print("  >> Updating individual with best performance...")
		for ind in self.population:
			if ind.fitness[0] > self.bestIndividual.fitness[0]:
				self.bestIndividual = ind
		if self.population[0] > self.bestIndividual:
			self.bestIndividual = self.population[0]
			#self.bestIndividual.fit(self.Tr_x, self.Tr_y)
			#self.bestIndividual.prun(min_dim = self.dim_min)


		#print("  >> Calculating number of dominated solutions...")
		[ind.getNumberOfDominatedSolutions(self.population) for ind in self.population]


		#print("  >> Calculating crownding distance...")
		frontiers = splitIntoFrontiers(self.population)
		for f in frontiers:
			for ind in f:
				ind.crowdingDistance = 0
			for objective in range(len(f[0].fitnesses)):
				f.sort(key= lambda x:x.fitness[objective], reverse=True)
				fmin = f[-1].fitness[objective]
				fmax = f[0].fitness[objective]
				if fmax-fmin>0.00001: # if this fitness is not constant
					f[0].crowdingDistance  = 99999
					f[-1].crowdingDistance = 99999
					for i in range(2, len(f)-1):
						f[i].crowdingDistance += \
							(f[i+1].fitness[objective]-f[i-1].fitness[objective])/(fmax-fmin)


		#print("  >> Sorting population using the number of dominated solutions...")
		self.population.sort(reverse=True)


		#print("  >> Generating Next Generation...")
		self.elite = getParetoFrontier(self.population)
		newPopulation = []

		while len(newPopulation) < self.population_size:
			offspring = getOffspring(self.rng, self.population, self.tournament_size)
			offspring = discardIndividualWithInvalidSize(offspring, self.dim_min, self.dim_max, self.max_depth)
			newPopulation.extend(offspring)
		self.old_population = self.population
		self.population = newPopulation[:self.population_size]

		end = time.time()

		#print("  >> Printing Progress...")
		if self.verbose and self.currentGeneration%1==0:
			if not self.Te_x is None:
				print("   > Gen #%2d:  Fitness: [ %s ] // Tr-Score: %.6f // Te-Score: %.6f  // Time: %.4f" % (self.currentGeneration, " - ".join("%.4f"%x for x in self.bestIndividual.getFitnesses()), self.bestIndividual.getTrainingMeasure(), self.bestIndividual.getTestMeasure(self.Te_x, self.Te_y), end- begin )  )
			else:
				print("   > Gen #%2d:  Fitness: %.6f // Tr-Score: %.6f // Time: %.4f" % (self.currentGeneration, self.bestIndividual.getFitnesses(),  self.bestIndividual.getTrainingMeasure(), end- begin )  )








	def predict(self, dataset):
		'''
		Returns the predictions for the samples in a dataset.
		'''
		self.checkIfTrained()

		return self.getBestIndividual().predict(dataset)

		return "Population Not Trained" if self.bestIndividual == None else self.bestIndividual.predict(sample)


def fitIndividuals(a):
	ind,x,y = a
	ind.getFitnesses(x,y)

	ret = []
	if "FOLD" in ind.fitnesses:
		ret.append(None)
	else:
		ret.append(ind.getTrainingPredictions())
	ret.append(ind.getFitnesses())
	ret.append(ind.model)

	
	return ret 









from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from .Node import Node
from sklearn.ensemble import RandomForestClassifier
from .MahalanobisDistanceClassifier import MahalanobisDistanceClassifier

import pandas as pd

from sklearn.metrics import accuracy_score, f1_score, cohen_kappa_score, mean_squared_error


# 
# By using this file, you are agreeing to this product's EULA
#
# This product can be obtained in https://github.com/jespb/Python-M6GP
#
# Copyright ©2023-2024 J. E. Batista
#


class Individual:
	training_X = None
	training_Y = None

	operators = None
	terminals = None
	max_depth = None

	dimensions: int = None
	size: int = 0
	depth: int = 0

	trainingPredictions = None
	testPredictions = None
	fitnesses = None
	fitness = None
	numberOfDominatedSolutions: int = -1

	model = None

	def __init__(self, operators, terminals, max_depth, model_name="RandomForestClassifier", fitnesses=["Accuracy","Size"]):
		self.operators = operators
		self.terminals = terminals
		self.max_depth = max_depth
		self.model_name = model_name
		self.fitnesses = fitnesses

	def create(self,rng, n_dims=1):
		self.dimensions = []
		for i in range(n_dims):
			n = Node()
			n.create(rng, self.operators, self.terminals, self.max_depth, full=True)
			self.dimensions.append(n)
		
	def copy(self, dim):
		self.dimensions = dim



	def __gt__(self, other):
		return self.numberOfDominatedSolutions >  other.numberOfDominatedSolutions

	def __ge__(self, other):
		return self.numberOfDominatedSolutions >= other.numberOfDominatedSolutions



	def __str__(self):
		return ",".join([str(d) for d in self.dimensions])


	def createModel(self):
		if self.model_name == "MahalanobisDistanceClassifier":
			return MahalanobisDistanceClassifier()
		elif self.model_name == "RandomForestClassifier":
			return RandomForestClassifier(random_state = 42, max_depth=4, n_estimators=5)
		elif self.model_name == "DecisionTreeClassifier":
			return DecisionTreeClassifier(random_state = 42, max_depth=4)
		elif self.model_name == "DecisionTreeRegressor":
			return DecisionTreeRegressor(random_state = 42, max_depth=6)

	def fit(self, Tr_x, Tr_y):
		'''
		Trains the classifier which will be used in the fitness function
		'''
		if self.model is None:
			self.training_X = Tr_x
			self.training_Y = Tr_y

			self.model = self.createModel()
	
			hyper_X = self.convert(Tr_x)

			self.model.fit(hyper_X,Tr_y)

	def getNumberOfDominatedSolutions(self, population):
		if self.numberOfDominatedSolutions == -1:
			count = 0 
			for other in population:
				dominates = True
				i = 0
				while i < len(self.fitnesses) and dominates:
					if self.fitness[i] <= other.fitness[i]:
						dominates = False
					i+=1
				if dominates:
					count += 1
			self.numberOfDominatedSolutions = count
		return self.numberOfDominatedSolutions



	def getSize(self):
		'''
		Returns the total number of nodes within an individual.
		'''
		if not self.size:
			self.size = sum(n.getSize() for n in self.dimensions)
		return self.size

	def getDepth(self):
		'''
		Returns the depth of individual.
		'''
		if not self.depth:
			self.depth = max([dimension.getDepth() for dimension in self.dimensions])
		return self.depth 

	def getDimensions(self):
		'''
		Returns a deep clone of the individual's list of dimensions.
		'''
		ret = []
		for dim in self.dimensions:
			ret.append(dim.clone())
		return ret


	def getNumberOfDimensions(self):
		'''
		Returns the total number of dimensions within the individual.
		'''
		return len(self.dimensions)



	def getFitnesses(self, tr_x = None, tr_y = None):
		'''
		Returns the individual's fitness.
		'''
		if self.fitness is None:
			if not tr_x is None:
				self.training_X = tr_x
			if not tr_y is None:
				self.training_Y = tr_y

			self.fitness=[]
			
			self.fit(self.training_X, self.training_Y)

			for fitnessType in self.fitnesses:
				if fitnessType == "Accuracy":
					self.getTrainingPredictions()
					acc = accuracy_score(self.trainingPredictions, self.training_Y)
					self.fitness.append(acc)

				if fitnessType == "MSE":
					self.getTrainingPredictions()
					mse = -1 * mean_squared_error(self.trainingPredictions, self.training_Y)
					self.fitness.append(mse)

				if fitnessType == "WAF":
					self.getTrainingPredictions()
					waf = f1_score(self.trainingPredictions, self.training_Y, average="weighted")
					self.fitness.append(waf)

				if fitnessType == "2FOLD":
					hyper_X = self.convert(self.training_X)

					X1 = hyper_X.iloc[:len(hyper_X)//2]
					Y1 = self.training_Y[:len(self.training_Y)//2]
					X2 = hyper_X.iloc[len(hyper_X)//2:]
					Y2 = self.training_Y[len(self.training_Y)//2:]

					M1 = self.createModel()
					M1.fit(X1,Y1)
					P1 = M1.predict(X2)

					M2 = self.createModel()
					M2.fit(X2,Y2)
					P2 = M2.predict(X1)

					f1 = accuracy_score(P1, Y2)
					f2 = accuracy_score(P2, Y1)
					self.fitness.append( (f1+f2)/2 )

				if fitnessType == "Size":
					self.fitness.append( 0 - self.getSize())

				if fitnessType == "Size(Classifier)":
					self.getTrainingPredictions()
					size = sum([x.tree_.node_count for x in self.model.estimators_ ])
					self.fitness.append( 0 - size )

				if fitnessType == "MDL":
					# accuracy
					self.getTrainingPredictions()
					acc = accuracy_score(self.trainingPredictions, self.training_Y)
					
					size_fe = self.getSize()
					size_classifier = sum([x.tree_.node_count for x in self.model.estimators_ ])

					self.fitness.append( 0- (size_fe + size_classifier + (1-acc)*len(self.training_Y)) )


		return self.fitness


	def getTrainingMeasure(self):
		if self.fitnesses[0] in ["Accuracy", "2FOLD"]:
			self.getTrainingPredictions()
			return accuracy_score(self.trainingPredictions, self.training_Y)
			
		if self.fitnesses[0] == "MSE":
			self.getTrainingPredictions()
			return -1 * mean_squared_error(self.trainingPredictions, self.training_Y)

		if self.fitnesses[0] == "WAF":
			self.getTrainingPredictions()
			return f1_score(self.trainingPredictions, self.training_Y, average="weighted")


	def getTestMeasure(self, test_X, test_Y):
		if self.fitnesses[0] in ["Accuracy", "2FOLD"]:
			self.getTestPredictions(test_X)
			return accuracy_score(self.testPredictions, test_Y)
			
		if self.fitnesses[0] == "MSE":
			self.getTestPredictions(test_X)
			return -1 * mean_squared_error(self.testPredictions, test_Y)

		if self.fitnesses[0] == "WAF":
			self.getTestPredictions(test_X)
			return f1_score(self.testPredictions, test_Y, average="weighted")



	def getTrainingPredictions(self):
		if self.trainingPredictions is None:
			self.trainingPredictions = self.predict(self.training_X)

		return self.trainingPredictions

	def getTestPredictions(self, X):
		if self.testPredictions is None:
			self.testPredictions = self.predict(X)

		return self.testPredictions


	
	def getMSE(self, X,Y,pred=None):
		'''
		Returns the individual's accuracy.
		'''
		if pred == "Tr":
			pred = self.getTrainingPredictions()
		elif pred == "Te":
			pred = self.getTestPredictions(X)
		else:
			pred = self.predict(X)

		return -1 * mean_squared_error(pred, Y)

	
	def getAccuracy(self, X,Y,pred=None):
		'''
		Returns the individual's accuracy.
		'''
		if pred == "Tr":
			pred = self.getTrainingPredictions()
		elif pred == "Te":
			pred = self.getTestPredictions(X)
		else:
			pred = self.predict(X)

		return accuracy_score(pred, Y)


	def getWaF(self, X, Y,pred=None):
		'''
		Returns the individual's WAF.
		'''
		if pred == "Tr":
			pred = self.getTrainingPredictions()
		elif pred == "Te":
			pred = self.getTestPredictions(X)
		else:
			pred = self.predict(X)

		return f1_score(pred, Y, average="weighted")


	def getKappa(self, X, Y,pred=None):
		'''
		Returns the individual's kappa value.
		'''
		if pred == "Tr":
			pred = self.getTrainingPredictions()
		elif pred == "Te":
			pred = self.getTestPredictions(X)
		else:
			pred = self.predict(X)

		return cohen_kappa_score(pred, Y)



	def calculate(self, sample):
		'''
		Return the position of a sample in the output space.
		'''
		return [self.dimensions[i].calculate(sample) for i in range(len(self.dimensions))]


	def convert(self, X):
		'''
		Returns the converted input space.
		'''
		ret = pd.DataFrame()
		for i in range(len(self.dimensions)):
			a = self.dimensions[i].calculate(X)
			ret["#"+str(i)] = a
		return ret


	def predict(self, X):
		'''
		Returns the class prediction of a sample.
		'''
		hyper_X = self.convert(X)
		predictions = self.model.predict(hyper_X)

		return predictions



	def prun(self,min_dim=1,simp=False):
		'''
		Remove the dimensions that degrade the fitness.
		If simp==True, also simplifies each dimension.
		'''

		dup = self.dimensions[:]
		i = 0
		ind = Individual(self.operators, self.terminals, self.max_depth, self.model_name, self.fitnesses)
		ind.copy(dup)

		ind.fit(self.training_X, self.training_Y)

		while i < len(dup) and len(dup) > min_dim:
			dup2 = dup[:]
			dup2.pop(i)
			ind2 = Individual(self.operators, self.terminals, self.max_depth, self.model_name, self.fitnesses)
			ind2.copy(dup2)
			ind2.fit(self.training_X, self.training_Y)

			if ind2 >= ind:
				ind = ind2
				dup = dup2
				i-=1
			i+=1
	
		self.dimensions = dup
		self.trainingAccuracy = None
		self.testAccuracy = None
		self.size = None
		self.depth = None
		self.model = None
		self.fit(self.training_X, self.training_Y)


		if simp:
			# Simplify dimensions
			for d in self.dimensions:
				done = False
				while not done:
					state = str(d)
					d.prun(self.training_X)
					done = state == str(d)



